//
// File:         sgi-midi.cc 
//
// Summary:      Midi tests for IRIX 5.3
//
// Author:       Tobias Kunze
// E-Mail:       t@kunze.stanford.edu
// Org:          CCRMA, Stanford University
//
// Orig-Date:    8-Jul-96 at 23:30:59
// Last-Mod:     16-Jul-96 at 22:48:39 by Tobias Kunze
//
// Description:   
// 
// Compile:      CC -g <name>.c -o <name>.o -lmd -ldmedia
//
// Changes:
//    00-Jan-00  tk    created

#ifdef _DEBUG_
#include <iostream.h>
#endif

#include <dmedia/midi.h>
#include <dmedia/dmedia.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <bstring.h>
#include <sys/time.h>

#ifndef TRUE
static const int TRUE = 1;
#endif
#ifndef FALSE
static const int FALSE = 0;
#endif

#ifndef Boolean
typedef unsigned char Boolean;
#endif

static int kSuccess = 0;
static int kFailure = -1;
//static int kChannelMessageTag = 0x4000000;
static int kPortBit = 0x8000;	// bit 15 used as a port bit

static MDport gLogicalInPort1;
static MDport gLogicalInPort2;
static MDport gLogicalOutPort1;
static MDport gLogicalOutPort2;

static int gLogicalIn1FD;
static int gLogicalIn2FD;

static char* gLogicalPort1Name = (char*) NULL;
static char* gLogicalPort2Name = (char*) NULL;
static char* gPhysicalPort1Name = (char*) NULL;
static char* gPhysicalPort2Name = (char*) NULL;
 
static void (*pLispReadHook)(long, long) = NULL;

fd_set gReadFDSet;
struct timeval gDummyTimeVal = {0L, 0L};


//static unsigned long gQueueTime = 0x7fffffffUL;
//static unsigned long gQueueTime = 0;
static int gTimeMode = MD_RELATIVESTAMP;
//static int gTimeMode = MD_DELTASTAMP;


//
// __MidiConnect
// -------------
// Create In- and OutPorts to the specified Interfaces.  
// In Irix 5.3, we have to use ticks and set the divisor manually to
// circumvent the nanosecond scheduling bugs.

inline void __InitPort(MDport port, int* desc) {
  //mdSetTemporalBuffering(gLogicalInPort1, gQueueTime); // no time limit
  //mdSetDivision(port, 25);
  //mdSetTempo(port, 25000);
  mdSetStampMode(port, gTimeMode);
  mdSetOrigin(port, 0);
  if (desc) {
    *desc = mdGetFd(port);
    FD_SET(*desc, &gReadFDSet);
  }
}

int __MidiConnect(char* lport1, char* lport2) {
  FD_ZERO(&gReadFDSet);
  if (lport1) {
    gLogicalInPort1 = mdOpenInPort(lport1);
    if (gLogicalInPort1 == (void*)NULL) {
#ifdef _DEBUG_
      cerr << "__MidiConnect(): couldn't create input connection to "
	   << lport1 << endl;
#else
      return -1;
#endif
    } else {
       __InitPort(gLogicalInPort1, &gLogicalIn1FD);
      gLogicalOutPort1 = mdOpenOutPort(lport1);
      if (gLogicalOutPort1 == (void*)NULL) {
#ifdef _DEBUG_
	cerr << "__MidiConnect(): couldn't create output connection to "
	     << lport1 << endl;
#else
	mdClosePort(gLogicalInPort1);	// failure: close gLogicalInPort1 again
	return -1;
#endif
      } else
	 __InitPort(gLogicalOutPort1, FALSE);
    }
  }
  if (lport2) {
    gLogicalInPort2 = mdOpenInPort(lport2);
    if (gLogicalInPort2 == (void*)NULL) {
#ifdef _DEBUG_
      cerr << "__MidiConnect(): couldn't create input connection to "
	   << lport2 << endl;
#else
      return -1;
#endif
    } else {
      __InitPort(gLogicalInPort2, &gLogicalIn2FD);
      gLogicalOutPort2 = mdOpenOutPort(lport2);
      if (gLogicalOutPort2 == (void*)NULL) {
#ifdef _DEBUG_
	cerr << "__MidiConnect(): couldn't create output connection to "
	     << lport2 << endl;
#else
	mdClosePort(gLogicalInPort2);	// failure: close gLogicalInPort2 again
	return -1;
#endif
      } else
	 __InitPort(gLogicalOutPort2, FALSE);
    }
  }
#ifdef _DEBUG_
  if (lport1)
  cout << "Log. Connection A to \"" << gLogicalPort1Name << "\": inport "
       << (void*)gLogicalInPort1 << " outport " << (void*)gLogicalOutPort1 << endl;
  if (lport2)
  cout << "Log. Connection B to \"" << gLogicalPort2Name << "\": inport "
       << (void*)gLogicalInPort2 << " outport " << (void*)gLogicalOutPort2 << endl;
#endif
  return kSuccess;
}


//
// MidiOpen
// --------
// Arguments:    mode:     -1  ports unspecified (do what seems reasonable)
//                          0  open port A
//                          1  open port B
//                          2  open port A and B
//                          3  open port B and A
//               readhook:  address of the lisp callback wrapper  
// Returns:      kSuccess or kFailure
// Description:  

int MidiOpen(int mode, void (*readhook)(long, long)) {
  int nports;

  nports = mdInit();
  switch (nports) {
  case -1 :
#ifdef _DEBUG_
    cerr << "MidiOpen(): Failure (no MIDI demon)" << endl;
#endif
    return kFailure;
  case 0 :
#ifdef _DEBUG_
    cerr << "MidiOpen(): Failure (no MIDI ports configured)" << endl;
#endif
    return kFailure;
  case 1 :
    gLogicalPort1Name = mdGetName(0);
    break;
  case 2 :
    gLogicalPort1Name = mdGetName(0);
    gLogicalPort2Name = mdGetName(1); 
    break;
  default :
#ifdef _DEBUG_
    cerr << "MidiOpen(): Can't handle mdInit() return value "
	 << nports << endl;
#else
    return kFailure;
#endif
  }

#ifdef _DEBUG_
  cout << "gLogicalPort1Name: " << (void*)gLogicalPort1Name << " \""
       << gLogicalPort1Name << "\"\n"
       << "gLogicalPort2Name: " << (void*)gLogicalPort2Name << " \""
       << gLogicalPort2Name << "\"\n";
#endif

  // determine physical device names.  I can't test the file descriptor's
  // device name with ttyname(3C) since this requires an already established
  // connection.  Here I simply check the whether the port names contain a `1'
  // or `2', respectively, and fall back to the original ordering, if not.

  int inorder = -2;		// -2 = unknown, -1 = undecidable, TRUE, FALSE

  if (gLogicalPort1Name) {
    if (strchr(gLogicalPort1Name, '1')) {
      gPhysicalPort1Name = gLogicalPort1Name;
      inorder = TRUE;
    } else if (strchr(gLogicalPort1Name, '2')) {
      gPhysicalPort2Name = gLogicalPort1Name;
      inorder = FALSE;
    } else
      inorder = -1;
  }

  if (gLogicalPort2Name && inorder != -1) {
    if (strchr(gLogicalPort2Name, '2')) {
      if (inorder != FALSE) {
	gPhysicalPort2Name = gLogicalPort2Name;
	inorder = TRUE;
      }
    } else if (strchr(gLogicalPort2Name, '1')) {
      if (inorder != TRUE) {
	gPhysicalPort1Name = gLogicalPort2Name;
	inorder = FALSE;
      }
    } else
      inorder = -1;
  }

  switch (inorder) {
  case -2 :			// unknown: should never occur
#ifdef _DEBUG_
    cerr << "MidiOpen(): can't determine port ordering (shouldn't)." << endl;
#endif
    break;
  case -1 :			// undecidable; treat as TRUE
  case TRUE :
    gPhysicalPort1Name = gLogicalPort1Name;
    gPhysicalPort2Name = gLogicalPort2Name;
    break;
  case FALSE :
    gPhysicalPort2Name = gLogicalPort1Name;
    gPhysicalPort1Name = gLogicalPort2Name;
    break;
  }

  // connect

#ifdef _DEBUG_
  if (-1 > mode || mode > 3)
    cerr << "MidiOpen(): Invalid mode " << mode
	 << ".  Making default connections." << endl;
#endif

  char* lport1;
  char* lport2;

  switch (mode) {
  case -1 :			// connect to whatever we have, do not warn
    lport1 = gLogicalPort1Name;
    lport2 = gLogicalPort2Name;
    break;
  case 0 :			// try first ttyd1, fallback to ttyd2
    lport1 = (gPhysicalPort1Name ? gPhysicalPort1Name : gPhysicalPort2Name);
    lport2 = (char*)NULL;
    break;
  case 1 :			// try first ttyd2, fallback to ttyd1
    lport1 = (gPhysicalPort2Name ? gPhysicalPort2Name : gPhysicalPort1Name);
    lport2 = (char*)NULL;
    break;
  case 3 : 			// try inverse, fallback to single port, warn
#ifdef _DEBUG_
    if ((! gLogicalPort1Name) || (! gLogicalPort2Name))
      cerr << "MidiOpen(): no second interface available." << endl;
#endif
    lport1 = (gPhysicalPort2Name ? gPhysicalPort2Name : gPhysicalPort1Name);
    lport2 = (gPhysicalPort2Name ?
	      (gPhysicalPort1Name ? gPhysicalPort1Name : (char*)NULL)
	      : (char*)NULL);
    break;
  case 2 :			// try normal, fallback to single port, warn
#ifdef _DEBUG_
    if ((! gLogicalPort1Name) || (! gLogicalPort2Name))
      cerr << "MidiOpen(): no second interface available." << endl;
#endif
    lport1 = (gPhysicalPort1Name ? gPhysicalPort1Name : gPhysicalPort2Name);
    lport2 = (gPhysicalPort1Name ?
	      (gPhysicalPort2Name ? gPhysicalPort2Name : (char*)NULL)
	      : (char*)NULL);
    break;
  }
  __MidiConnect(lport1, lport2);

  pLispReadHook = readhook;

  return kSuccess;
}


//
// MidiClose
// ---------
// No error checking here.

int MidiClose() {
  if (gLogicalInPort1) mdClosePort(gLogicalInPort1);
  if (gLogicalInPort2) mdClosePort(gLogicalInPort2);
  if (gLogicalOutPort1) mdClosePort(gLogicalOutPort1);
  if (gLogicalOutPort2) mdClosePort(gLogicalOutPort2);
  return kSuccess;
}



//
// MidiSetTime
// -----------
// Set the time to UST (Unadjusted System Time)

int MidiSetTime(long ms) {
  unsigned long long ns;
  dmGetUST(&ns);
  ns = ns - (ms * NSEC_PER_MSEC); // ms to ns
  if (gLogicalInPort1) mdSetOrigin(gLogicalInPort1, ns);
  if (gLogicalOutPort1) mdSetOrigin(gLogicalOutPort1, ns);
  if (gLogicalInPort2) mdSetOrigin(gLogicalInPort2, ns);
  if (gLogicalOutPort2) mdSetOrigin(gLogicalOutPort2, ns);
  return kSuccess;
}


// 
// MidiGetTime
// -----------

long MidiGetTime() {
  unsigned long long ust;
  dmGetUST(&ust);
  // all ports have the same origin, so get it from only one.
  unsigned long long orig = mdGetOrigin((gLogicalInPort1 ? gLogicalInPort1 : gLogicalInPort2));
  return (long)((ust - orig)/NSEC_PER_MSEC);
}


//
// MidiStartTime
// -------------
// Equivalent to MidiSetTime(0);

int MidiStartTime() {
  MidiSetTime(0);
  return kSuccess;
}


// 
// MidiStopTime
// ------------
// Dummy.  Noone will ever need such a function.

int MidiStopTime() {
  return kSuccess;
}



//
// Midi Reading and Writing
//

MDevent gEvent;

//
// MidiReadMessages
// ----------------

inline void __CallLisp() {
  // do only channel messages
  if (0x80 <= gEvent.msg[0] && gEvent.msg[0] <= 0xEf) { 
#ifdef _TESTING_
    char printbuf[80];
    mdPrintEvent(printbuf, &gEvent, 1);
    cout << printbuf;
#else
    // encode message
    long msg = 0L;
    msg |= 0x7000000;		// channel message, 3 bytes long
    msg |= gEvent.msg[0] << 16;
    msg |= gEvent.msg[1] << 8;
    msg |= gEvent.msg[2];
    (*pLispReadHook)(msg, (long)(gEvent.stamp / NSEC_PER_MSEC));
#endif
  }
}  



int MidiReadMessages() {
  if (gLogicalInPort1 && gLogicalInPort2) {
    while (1) {
      FD_SET(gLogicalIn1FD, &gReadFDSet);
      FD_SET(gLogicalIn2FD, &gReadFDSet);
      if (select(2, &gReadFDSet, (fd_set*)NULL, (fd_set*)NULL,
		 &gDummyTimeVal) == 0)
	break;
      else {
	if (FD_ISSET(gLogicalIn1FD, &gReadFDSet)) {
	  mdReceive(gLogicalInPort1, &gEvent, 1);
	  __CallLisp();
	}
	if (FD_ISSET(gLogicalIn2FD, &gReadFDSet)) {
	  mdReceive(gLogicalInPort2, &gEvent, 1);
	  gEvent.msg[1] |= 0x80;	// set the port bit
	  __CallLisp();
	}
      }
    }
    return kSuccess;
  } else if (gLogicalInPort1) {
#ifdef _DEBUG_
    cout << "Reading from port A, FD " << gLogicalIn1FD << " "
	 << &gReadFDSet << endl;
#endif
    int res;
    while (1) {
      FD_ZERO(&gReadFDSet);
#ifdef _DEBUG_
      cout << (long*)&gReadFDSet << " " << (long*)(&gReadFDSet + 4) << " "
	   << (long*)(&gReadFDSet + 8) << " " << (long*)(&gReadFDSet + 12) << " "
	   << (long*)(&gReadFDSet + 16) << " " << (long*)(&gReadFDSet + 20) << " "
	   << (long*)(&gReadFDSet + 24) << " " << (long*)(&gReadFDSet + 28) << " "
	   << endl;
#endif
      FD_SET(gLogicalIn1FD, &gReadFDSet);
#ifdef _DEBUG_
      cout << (long*)&gReadFDSet << " " << (long*)(&gReadFDSet + 4) << " "
	   << (long*)(&gReadFDSet + 8) << " " << (long*)(&gReadFDSet + 12) << " "
	   << (long*)(&gReadFDSet + 16) << " " << (long*)(&gReadFDSet + 20) << " "
	   << (long*)(&gReadFDSet + 24) << " " << (long*)(&gReadFDSet + 28) << " "
	   << endl;
#endif
      res = select(1, &gReadFDSet, (fd_set*)NULL, (fd_set*)NULL,
		   &gDummyTimeVal);
#ifdef _DEBUG_
      cout << res << " " << FD_ISSET(gLogicalIn1FD, &gReadFDSet) << endl;
#endif
      if (res == 0)
	break;
      else {
	mdReceive(gLogicalInPort1, &gEvent, 1);
      	__CallLisp();
      }
    }
    return kSuccess;
  } else {
    while (1) {
      FD_SET(gLogicalIn2FD, &gReadFDSet);
      if (select(1, &gReadFDSet, (fd_set*)NULL, (fd_set*)NULL,
		 &gDummyTimeVal) == 0)
	break;
      else {
	mdReceive(gLogicalInPort2, &gEvent, 1);
	gEvent.msg[1] |= 0x80;	// set the port bit
      	__CallLisp();
      }
    }
    return kSuccess;
  }
}



// 
// MidiWriteMessage
// ----------------
// This only deals with channel messages.  No meta-, sytem-, or raw data
// messages are being sent or received so far.
// Time is assumed to come in in milliseconds and gets converted to
// nanoseconds.

int MidiWriteMessage(long msg, long tim) {
#ifdef _DEBUG_
  cout << "Timestamp: " << (tim * NSEC_PER_MSEC) << endl;
#endif
  gEvent.stamp = (unsigned long long)(tim * NSEC_PER_MSEC); // ms to ns
  gEvent.msglen = (msg & 0x03000000) >> 24;
  gEvent.msg[0] = (msg & 0xff0000) >> 16;
  gEvent.msg[1] = (msg & 0xff00) >> 8;
  gEvent.msg[2] = msg & 0xff;

  if ((msg & kPortBit)) {
    if (gLogicalOutPort2)
      if (mdSend(gLogicalOutPort2, &gEvent, 1) == -1) {
#ifdef _DEBUG_
	cout << "mdSend(): failed on gLogicalOutPort1." << endl;
#endif
	return kFailure;
      }
  } else {
    if (gLogicalOutPort1)
      if (mdSend(gLogicalOutPort1, &gEvent, 1) == -1){
#ifdef _DEBUG_
	cout << "mdSend(): failed on gLogicalOutPort2." << endl;
#endif
      }
  }
  return kSuccess;
}


//
// MidiWriteNote
// -------------
// Specialized output function for midi-notes.  Takes a note-on messge
// and its queue-time plus a duration.
 
int MidiWriteNote(long msg, long time, long dur) {
#ifdef _DEBUG_
  cout << "Writing Note-On:  " << (void*)msg << " @ " << time
       << "\n        Note-Off: " << (void*)(msg & 0xff8fffff) << " @ "
       << time + dur << endl;
#endif
  MidiWriteMessage(msg, time);
  MidiWriteMessage(msg & 0xff8fffff, time + dur);
  return kSuccess;
}
 


//
// Flushing is implemented in the Schwarzenegger fashion, i.e. by shutting
// ports down and re-opening them immediately.

//
// MidiFlushTransmit
// -----------------

int MidiFlushTransmit() {
  Boolean outportA = (gLogicalOutPort1 ? TRUE : FALSE);
  Boolean outportB = (gLogicalOutPort2 ? TRUE : FALSE);
  
  if (outportA) mdClosePort(gLogicalOutPort1);
  if (outportB) mdClosePort(gLogicalOutPort2);
  
  if (outportA) {
    gLogicalOutPort1 = mdOpenOutPort(gLogicalPort1Name);
    if (gLogicalOutPort1 == NULL) {
#ifdef _DEBUG_
      cerr << "MidiFlushTransmit(): couldn't re-create output connection to "
	   << gLogicalPort1Name << endl;
#else
    ;
#endif
    } else {
      __InitPort(gLogicalOutPort1, FALSE);
      mdPanic(gLogicalOutPort1);
    }
  }
  if (outportB) {
    gLogicalOutPort2 = mdOpenOutPort(gLogicalPort2Name);
    if (gLogicalOutPort2 == NULL) {
#ifdef _DEBUG_
      cerr << "MidiFlushTransmit(): couldn't re-create output connection to "
	   << gLogicalPort2Name << endl;
#else 
      ;
#endif
    } else {
      __InitPort(gLogicalOutPort2, FALSE);
      mdPanic(gLogicalOutPort2);
    }
  }
  return kSuccess;
}




// 
// MidiFlushReceive
// ----------------
// Dummy.  Not implemented in dmedia, it seems.

int MidiFlushReceive() {
  Boolean inportA = (gLogicalInPort1 ? TRUE : FALSE);
  Boolean inportB = (gLogicalInPort2 ? TRUE : FALSE);

  if (inportA) mdClosePort(gLogicalInPort1);
  if (inportB) mdClosePort(gLogicalInPort2);

  FD_ZERO(&gReadFDSet);
  if (inportA) {
    gLogicalInPort1 = mdOpenInPort(gLogicalPort1Name);
    if (gLogicalInPort1 == NULL) {
#ifdef _DEBUG_
      cerr << "MidiFlushReceive(): couldn't re-create input connection from "
	   << gLogicalPort1Name << endl;
#else 
      ;
#endif
    } else
      __InitPort(gLogicalInPort1, &gLogicalIn1FD);
  }
  if (inportB) {
    gLogicalInPort2 = mdOpenInPort(gLogicalPort2Name);
    if (gLogicalInPort2 == NULL) {
#ifdef _DEBUG_
      cerr << "MidiFlushReceive(): couldn't re-create input connection from "
	   << gLogicalPort2Name << endl;
#else
      ;
#endif
    } else
      __InitPort(gLogicalInPort2, &gLogicalIn2FD);
  }

  return kSuccess;
}



// 
// MidiHush
// --------
// Just a combination of MidiFlushReceive() and MidiFlushTransmit().

int MidiHush() {
  MidiFlushTransmit();
  MidiFlushReceive();
  return kSuccess;
}



//
// MidiAllNotesOff
// ---------------
// The argument time is ignored.

int MidiAllNotesOff(long time) {
  long dummy = time;		// make the compiler happy
  dummy += 1;
  if (gLogicalOutPort1) mdPanic(gLogicalOutPort1);
  if (gLogicalOutPort2) mdPanic(gLogicalOutPort2);
  return kSuccess;
}



//--------------------------------------------------------------------------

#ifdef _TESTING_

//
// main
// ----

void main() {
  // create midi ports and connect them to the interface(s)
  MidiOpen(2, (void (*)(long, long))0);

  long curTime = MidiGetTime();
  cout << "CurTime: " << curTime << endl;

  unsigned char i;
  int n;

  for (i = 60, n = 0; i <= 72; i++, n += 100) {
    MidiWriteNote((0x7900064 | (i << 8)), curTime + n, 100);
  }
  for (i = 1; i < 10; i++) {
    MidiReadMessages();
    sleep(1);
  }
  MidiClose();

  cout << endl;  
}

#endif // _TESTING_

//
// -*- EOF -*-

